﻿using System;
using System.IO;
using System.Collections.Generic;
using FinMath.ClusterAnalysis;
using FinMath.LinearAlgebra;
using FinMath.DataStructures;

namespace FinMath.Tests.Performance
{
	public static class ClusterAnalysis
	{
		#region Tests Settings

		// Following arrays must have the same size.
		static Int64[] objectsCount = { 
				500, 500, 500, 500, 500, 
				1000, 1000, 1000, 1000, 1000, 
				1500, 1500, 1500, 1500, 1500 };
		static Int64[] factorsCount = { 
				200, 350, 500, 1000, 1500, 
				200, 350, 500, 1000, 1500, 
				200, 350, 500, 1000, 1500};
		static Int64[] repeatsCount = { 
				600, 400, 300, 150, 100, 
				150, 100, 70, 40, 20, 
				70, 40, 30, 12, 7 };

		// FinMath metric names.
		// If you change this you must also do appropriate changes in matlab file.
		static MetricType[] finmaths = {
			MetricType.CityblockMetric, 
			MetricType.EuclideanDistance, 
			MetricType.SquaredEuclideanDistance, 
			MetricType.ChebychevDistance, 
			MetricType.CosineSimilarity, 
			MetricType.HammingMetric,
			MetricType.CorrelationDistance};
		// Experiments parameters.
		#endregion

		static Int32 lineLength = 69;

		#region FinMath Computations
		public static TimeSpan ClusterizeFinMath(Matrix X, int repeatCount, MetricType type)
		{
			SparseAgglomerativeClustering sc = new SparseAgglomerativeClustering(1);

			DateTime before = DateTime.Now;

			for (int i = 0; i < repeatCount; ++i)
			{
				Matrix distance = MetricsCalculator.CalculateMetric(X, type);
				sc.ComputeClustering(distance);
			}

			return (DateTime.Now - before);
		}
		#endregion

		#region Console Utils

		public static string FormatType(MetricType type)
		{
			String caption = type.ToString();
			Int32 l = lineLength - caption.Length - 4;
			String left = "", right;
			for (Int32 i = 0; i < l / 2; ++i)
				left += ' ';
			right = (string)left.Clone();
			if (l % 2 == 1)
				right += ' ';
			return "| " + left + caption + right + " |";
		}

		public static void DrawLine(Int32 charCount)
		{
			String line = "";
			for (int i = 0; i < charCount; ++i)
				line += "_";

			Console.WriteLine(line);
		}

		public static void DrawHeader(StreamWriter writer, MetricType caption)
		{
			DrawLine(lineLength);
			Console.WriteLine(FormatType(caption));
			DrawLine(lineLength);
			Console.Write("| Objects  |  Factors | Repeats |");
			Console.WriteLine("   Total   |  Average  |  Single   |");
			Console.Write("|  count   |   count  |  count  |");
			Console.WriteLine("   Time    |  MFLOPS   | Calc Time |");
			DrawLine(lineLength);

			writer.WriteLine(caption);
			writer.WriteLine("Objects,Factors,Repeats,Matlab Time,FinMath MFLOPS,Single calc time");
		}

		#endregion

		static void CreateTest()
		{
			if (Directory.Exists("..\\..\\Tests"))
				Directory.Delete("..\\..\\Tests", true);
			Directory.Delete("..\\..\\Tests", true);
			Directory.CreateDirectory("..\\..\\Tests");
			for (int i = 0; i < objectsCount.Length; ++i)
			{
				Matrix X = Matrix.Random((Int32)factorsCount[i], (Int32)objectsCount[i]);
				X.Save("..\\..\\Tests\\" + i.ToString("00") + "." + objectsCount[i] + "." + factorsCount[i] + "." + repeatsCount[i] + ".csv");
			}
		}

		static void RunTest()
		{
			DirectoryInfo dir = new DirectoryInfo("..\\..\\Tests");
			FileInfo[] tests = dir.GetFiles("*.csv");
			StreamWriter writer = new StreamWriter("..\\..\\finmath-results.csv");

			foreach (MetricType metric in finmaths)
			{
				DrawHeader(writer, metric);

				foreach (FileInfo test in tests)
				{
					Matrix X = Matrix.Load(test.FullName);
					String[] parameters = test.Name.Split('.');
					Int64 o = Int64.Parse(parameters[1]);
					Int64 f = Int64.Parse(parameters[2]);
					Int64 r = Int64.Parse(parameters[3]);

					TimeSpan s = ClusterizeFinMath(X, (Int32)r, metric);

					Double operationsCount = o * o * f * r;
					Double t = s.TotalMilliseconds / 1e+3;
					Double p = operationsCount / t / 1e+6;
					Double it = t / r;

					Console.Write(
						"| {0,8:D} | {1,8:D} | {2,7:D} | ",
						o,
						f,
						r);

					Console.WriteLine(
						"{0,9} | {1,9} | {2,9} |",
						t.ToString("00.000"),
						Math.Ceiling(p),
						it.ToString("00.000"));
					DrawLine(lineLength);

					writer.WriteLine(
						"{0},{1},{2},{3},{4},{5}",
						o,
						f,
						r,
						t.ToString("00.000"),
						Math.Ceiling(p),
						it.ToString("00.000"));
					writer.Flush();
					
				}

				Console.WriteLine();
			}
		}

		static void Main(string[] arguments)
		{
			CreateTest();
			RunTest();
		}
	}
}
